﻿#include <iostream>
#include <array>
#include <vector>
#include <random>
#include <chrono>
#include <string>
#include <type_traits>
#include <omp.h>
//#include <gmpxx.h>
#include <boost/multiprecision/cpp_int.hpp>
#include "Montgomery.cpp"
#include "Poly.cpp"
#include "Matrix.cpp"
#include "PrintInt128.cpp"
#include "isaac/isaac64.c"
#include "RandomDist.cpp"
#include "Setup.cpp"
#include "SIMD.cpp"

using namespace std;

typedef vector<Matrix<Poly<d, Z>>> Mvector;

struct FHEKey {
    vector<Z> q;
    Mvector s;
    Mvector A;
    Mvector tau;
};

Matrix<Poly<d, Z>> SecretKeyGen(int n);
Matrix<Poly<d, Z>> PublicKeyGen(Matrix<Poly<d, Z>>& s, Z q, int N);
Matrix<Poly<d, Z>> Enc(Matrix<Poly<d, Z>>& A, Poly<d, Z>& m);
Matrix<Poly<d, Z>> Enc(Poly<d, Z>& m, FHEKey key, int index);
Matrix<Poly<d, Z>> EncZero(FHEKey key, int index);
Poly<d, Z> Dec(Matrix<Poly<d, Z>> c, Matrix<Poly<d, Z>> s, Z q);
Poly<d, Z> Dec(Matrix<Poly<d, Z>> c, FHEKey& key, int index);
Matrix<Poly<d, Z>> BitDecomp(Matrix<Poly<d, Z>> x, Z q);
Matrix<Poly<d, Z>> PowersOf2(Matrix<Poly<d, Z>> x, Z q);
Matrix<Poly<d, Z>> SwitchKeyGen(Matrix<Poly<d, Z>> s1, Matrix<Poly<d, Z>> s2, Z q);
Matrix<Poly<d, Z>> SwitchKey(Matrix<Poly<d, Z>> c, Matrix<Poly<d, Z>> B, Z q);
Matrix<Poly<d, Z>> Scale(Matrix<Poly<d, Z>> x, Z q, Z p, int r);

double Error(Matrix<Poly<d, Z>> c, Matrix<Poly<d, Z>> s, Z q);

Mvector EncryptBuffer(Matrix<Poly<d, Z>>& A, uint8_t* data, size_t length, int blocksize);
Mvector EncryptInteger(Matrix<Poly<d, Z>>& A, Z n);
Mvector EncryptInteger(Z n, FHEKey& key, int index);
Mvector EncryptIntegerPack(vector<int> pack, FHEKey& key, int index);
Mvector EncryptString(string str, FHEKey& key, int index);
Z DecryptInteger(Mvector c, Matrix<Poly<d, Z>>& s, Z q);
Z DecryptInteger(Mvector c, FHEKey& key, int index);
vector<int> DecryptIntegerPack(Mvector c, FHEKey& key, int index);
Mvector SumIntegersZ2(Mvector& c1, Mvector& c2, FHEKey& key, int index);
int SumIntegersLogZ2(Mvector c1, Mvector c2, Mvector& out, FHEKey& key, int index);
void ThreeForTwoZ2(Mvector c1, Mvector c2, Mvector c3, Mvector& sumOut, Mvector& carryOut, FHEKey& key, int index);
int ThreeForTwoZ2(vector<Mvector> numbers, Mvector& sumOut, Mvector& carryOut, FHEKey& key, int index);
Matrix<Poly<d, Z>> Search(Mvector data, Mvector word, FHEKey& key, int& index);

void FHEKeyGen(FHEKey& key);
Matrix<Poly<d, Z>> Refresh(Matrix<Poly<d, Z>> c, FHEKey& key, int index);
Mvector Refresh(Mvector c, FHEKey& key, int index);

bool MillerRabin(Z p);
Z firstPrimeAbove(Z n);
Z firstPrimeAboveWithConditions(Z n);
Z gcdExtended(Z a, Z b, Z& x, Z& y);

int main()
{   
    random_device r;
    for (int i = 0; i < RANDSIZ; i++) randrsl[i] = r();
    randinit(TRUE);
    InitBatching();

    auto t1 = std::chrono::high_resolution_clock::now();
    auto t2 = std::chrono::high_resolution_clock::now();

    uZ q = 1;
    q <<= 125;

    t1 = std::chrono::high_resolution_clock::now();
    FHEKey key;
    for (int i = 0; i < 12; i++) {
        key.q.push_back(firstPrimeAboveWithConditions(q));
        cout << key.q[i] << endl;
        q /= 800;
    }

    //Poly<d, Z>::push_q(key.q[2]);
    //InitBatching();
    //vector<Poly<d, Z>> moduli = {
    //    {1,1,0,1},
    //    {1},
    //    {0,1}
    //};
    //cout << Batch(moduli) << endl;
    //Poly<d, Z> F1 = { 1, 0, 0, 0, 1, 1, 0, 0, 1, 0, 0, 1, 1, 0, 0, 0, 1 };
    //cout << Unbatch(Batch(moduli))[0] << endl;
    
    FHEKeyGen(key);
    t2 = std::chrono::high_resolution_clock::now();
    cout << "Clau generada en " << (double)chrono::duration_cast<std::chrono::milliseconds>(t2 - t1).count() / 1000 << " s" << endl;

    Poly<d, Z> m1 = { 1 , 0, 1, 1, 0, 0, 1 };
    Poly<d, Z> m2 = { 0 , 1, 0, 0, 0, 0, 1 };
    Poly<d, Z> m3 = { 1 , 1 };
    cout << (m1 * m2) % q << endl;
    Poly<d, Z>::push_q(key.q[0]);

    //cout << (A[0] * s[0]) % key.q[0] << endl;
    auto c1 = Enc(key.A[0], m3);
    cout << Dec(c1, key.s[0], key.q[0]) << endl;
    cout << Error(c1, key.s[0], key.q[0]) << endl;

    for (int i = 0; i < 10; i++) {
        auto c2 = c1.outer(c1);
        c1 = Refresh(c2, key, i);
        Poly<d, Z>::push_q(key.q[i + 1]);
        cout << Dec(c1, key.s[i + 1], key.q[i + 1]) << endl;
        cout << Error(c1, key.s[i + 1], key.q[i + 1]) << endl << endl;
    }
    
    Mvector str1 = EncryptString("Una introducció a la criptografia homomòrfica. Implementació de l'esquema BGV.", key, 0);
    Mvector word = EncryptString("BGV", key, 0);
    int index = 0;
    t1 = std::chrono::high_resolution_clock::now();
    auto exists = Search(str1, word, key, index);
    t2 = std::chrono::high_resolution_clock::now();
    cout << "Cerca realitzada en " << (double)chrono::duration_cast<std::chrono::milliseconds>(t2 - t1).count() / 1000 << " s" << endl;
    cout << Dec(exists, key, index);
    cout << index << endl;
    //return 0;

    vector<int> notes1 = { 2, 7, 8, 9, 2, 1, 5, 8, 2, 2, 1, 0, 0, 5, 10, 10 };
    vector<int> notes2 = { 5, 9, 5, 8, 2, 7, 2, 7, 8, 7, 5, 9, 0, 1, 3, 3 };
    vector<int> notes3 = { 4, 7, 6, 6, 7, 7, 7, 7, 8, 8, 8, 10, 0, 1, 1, 1 };
    vector<int> notes4 = { 4, 2, 5, 9, 2 , 5, 6, 7, 8, 9, 10, 10, 2, 3, 1, 0};

    vector<Mvector> IntegersPack = { EncryptIntegerPack(notes1, key, 0), EncryptIntegerPack(notes2, key, 0), EncryptIntegerPack(notes3, key, 0), EncryptIntegerPack(notes4, key, 0) };
    cout << DecryptIntegerPack(IntegersPack[0], key, 0)[0] << endl;
    Mvector sum(8), carry(9);
    int iter = ThreeForTwoZ2(IntegersPack, sum, carry, key, 0);
    iter = SumIntegersLogZ2(sum, carry, sum, key, 0 + iter);
    auto integers = DecryptIntegerPack(sum, key, iter);
    for (int i = 0; i < integers.size(); i++) {
        cout << integers[i] << " ";
    }
    cout << endl;
    cout << iter << endl;

    //vector<Mvector>Integers = { EncryptInteger(100, key, 0)
    //    , EncryptInteger(13, key, 0), EncryptInteger(19, key, 0), EncryptInteger(10, key, 0), EncryptInteger(1, key, 0) };
    //Mvector sum(8), carry(9);
    //int iter = ThreeForTwoZ2(Integers, sum, carry, key, 0);
    //iter = SumIntegersLogZ2(sum, carry, sum, key, 0 + iter);
    //cout << DecryptInteger(sum, key, iter) << endl;
    //cout << iter << endl;

    //auto t1 = std::chrono::high_resolution_clock::now();

    //auto t2 = std::chrono::high_resolution_clock::now();
    //cout << chrono::duration_cast<std::chrono::milliseconds>(t2 - t1).count() << endl;

    cout << Montgomery::counter << endl;
    return 0;
}

Mvector EncryptBuffer(Matrix<Poly<d, Z>>& A, uint8_t* data, size_t length, int blocksize) {
    Mvector res;
    Poly<d, Z> m;
    for (int i = 0, j = 0; i < length*8; i++, j++) {
        m[j] = (data[i / 8] >> (i % 8)) & 1;
        if (j == blocksize - 1 || i == length*8 - 1) {
            res.push_back(Enc(A, m));
            j = -1;
        }
    }
    return res;
}

Mvector EncryptInteger(Matrix<Poly<d, Z>>& A, Z n) {
    return EncryptBuffer(A, (uint8_t*)&n, 1, 1);
}

Mvector EncryptInteger(Z n, FHEKey& key, int index) {
    Poly<d, Z>::push_q(key.q[index]);
    auto c = EncryptBuffer(key.A[index], (uint8_t*)&n, 1, 1);
    Poly<d, Z>::pop_q();
    return c;
}

Mvector EncryptIntegerPack(vector<int> pack, FHEKey& key, int index) {
    Mvector res;
    Poly<d, Z>::push_q(key.q[index]);
    for (int i = 0; i < 8; i++) {
        vector<Poly<d, Z>> bits;
        for (int j = 0; j < pack.size(); j++) {
            bits.push_back({ (pack[j] >> i) & 1 });
        }
        Poly<d, Z> m = Batch(bits);
        res.push_back(Enc(m, key, index));
    }
    Poly<d, Z>::pop_q();
    return res;
}

Mvector EncryptString(string str, FHEKey& key, int index) {
    Poly<d, Z>::push_q(key.q[index]);
    auto c = EncryptBuffer(key.A[index], (uint8_t*)str.data(), str.length(), 8);
    Poly<d, Z>::pop_q();
    return c;
}

Z DecryptInteger(Mvector c, Matrix<Poly<d, Z>>& s, Z q) {
    Z n = 0;
    for (int i = c.size() - 1; i >= 0; i--) {
        n <<= 1;
        n += Dec(c[i], s, q)[0];
    }
    return n;
}

Z DecryptInteger(Mvector c, FHEKey& key, int index) {
    return DecryptInteger(c, key.s[index], key.q[index]);
}

vector<int> DecryptIntegerPack(Mvector c, FHEKey& key, int index) {
    vector<int> integers(16);
    for (int i = c.size() - 1; i >= 0; i--) {
        //cout << c[i] << endl;
        Poly<d, Z> m = Dec(c[i], key, index);
        auto bits = Unbatch(m);
        //cout << bits[0] << endl;
        for (int j = 0; j < bits.size(); j++) {
            integers[j] <<= 1;
            integers[j] += (int)bits[j][0];
        }
    }
    return integers;
}

Mvector SumIntegersZ2(Mvector& c1, Mvector& c2, FHEKey& key, int index) {
    int n = c1.size();
    cout << c1.size() << " " << c2.size() << endl;
    Mvector res(n);
    Poly<d, Z>::push_q(key.q[index]);
    res[0] = Refresh(c1[0] + c2[0], key, index);
    auto carry = Refresh(c1[0].outer(c2[0]), key, index);
    //cout << Dec(c1[0], key.s[index], key.q[index]) << endl;
    //cout << Dec(c2[0], key.s[index], key.q[index]) << endl;
    //cout << Dec(res[0], key.s[index + 1], key.q[index + 1]) << endl;
    //cout << Dec(carry, key.s[index + 1], key.q[index + 1]) << endl << endl;
    for (int i = 1; i < n; i++) {
        auto c1_q = c1[i], c2_q = c2[i];
        for (int j = 0; j < i; j++) {
            c1_q = Refresh(c1_q, key, index + j);
            c2_q = Refresh(c2_q, key, index + j);
        }
        Poly<d, Z>::push_q(key.q[index + i]);
        res[i] = Refresh(c1_q + c2_q + carry, key, index + i);
        carry = Refresh(c1_q.outer(c2_q) + c1_q.outer(carry) + c2_q.outer(carry), key, index + i);

        //cout << Dec(c1_q, key.s[index + i], key.q[index + i]) << endl;
        //cout << Dec(c2_q, key.s[index + i], key.q[index + i]) << endl;
        //cout << Dec(res[i], key.s[index + i +1], key.q[index + i +1]) << endl;
        //cout << Dec(carry, key.s[index + i +1], key.q[index + i +1]) << endl << endl;
    }
    for (int i = 0; i < n - 1; i++) {
        for(int j = 1; j < n - i; j++) res[i] = Refresh(res[i], key, index + i + j);
    }
    return res;
}

int SumIntegersLogZ2(Mvector c1, Mvector c2, Mvector& out, FHEKey& key, int index) {
    int n = c1.size();
    int N = log2(n);
    cout << n << " " << N << endl;
    int indexInit = index;
    vector<Mvector> G, P, C;
    Mvector Gi(n), Pi(n), Ci(n), a(n), b(n);
    Poly<d, Z>::push_q(key.q[index]);
    for (int i = 0; i < n; i++) {
        Gi[i] = Refresh(c1[i].outer(c2[i]), key, index);
        Pi[i] = Refresh(c1[i] + c2[i], key, index);
    }
    a = Refresh(c1, key, index);
    b = Refresh(c2, key, index);
    index++;
    Poly<d, Z>::pop_q();
    G.push_back(Gi);
    P.push_back(Pi);
    for (int k = 1; k <= N; k++) {
        Poly<d, Z>::push_q(key.q[index]);
        Gi.clear();
        Pi.clear();
        for (int i = 0; i < (n >> k); i++) {
            //cout << "k=" << k << " i=" << i << endl;
            Gi.push_back(Refresh(G[k - 1][2 * i].outer(P[k - 1][2 * i + 1]).addToFirstRow(G[k - 1][2 * i + 1]), key, index));
            Pi.push_back(Refresh(P[k - 1][2 * i].outer(P[k - 1][2 * i + 1]), key, index));
            //cout << Dec(Pi[i], key.s[index + 1], key.q[index + 1]) << endl;
        }
        G.push_back(Gi);
        P.push_back(Pi);
        a = Refresh(a, key, index);
        b = Refresh(b, key, index);
        Poly<d, Z>::pop_q();
        index++;
    }
    C.push_back(G.back());

    for (int k = 0; k < N; k++) {
        Poly<d, Z>::push_q(key.q[index]);
        Pi = P[N - k - 1];
        Gi = G[N - k - 1];
        for (int j = indexInit + N - k; j < index; j++) {
            Pi = Refresh(Pi, key, j);
            Gi = Refresh(Gi, key, j);
        }
        for (int i = 0; i < (1 << k); i++) {
            //cout << "k=" << k << " i=" << i << endl;
            Ci[2 * i + 1] = Refresh(C[k][i].outer(Pi[2 * i]).addToFirstRow(Gi[2 * i]), key, index);
            cout << Dec(Ci[2 * i + 1], key.s[index + 1], key.q[index + 1]) << endl;
            Ci[2 * i] = Refresh(C[k][i], key, index);
            //cout << Dec(Ci[2*i+1], key.s[index + 1], key.q[index + 1]) << endl;
        }
        C.push_back(Ci);
        a = Refresh(a, key, index);
        b = Refresh(b, key, index);
        Poly<d, Z>::pop_q();
        index++;
    }
    Poly<d, Z>::push_q(key.q[index]);
    Mvector res(n);
    for (int i = 0; i < n; i++) {
        res[i] = Refresh(a[i] + b[i] + C[N][i], key, index);
    }
    Poly<d, Z>::pop_q();
    index++;
    out = res;
    return index;
}

void ThreeForTwoZ2(Mvector c1, Mvector c2, Mvector c3, Mvector& sumOut, Mvector& carryOut, FHEKey& key, int index) {
    int n = c1.size();
    carryOut[0] = EncZero(key, index + 1);
    Poly<d, Z>::push_q(key.q[index]);
    for (int i = 0; i < n; i++) {
        sumOut[i] = Refresh(c1[i] + c2[i] + c3[i], key, index);
        carryOut[i + 1] = Refresh(c1[i].outer(c2[i]) + c1[i].outer(c3[i]) + c2[i].outer(c3[i]), key, index);
    }
    Poly<d, Z>::pop_q();
}

int ThreeForTwoZ2(vector<Mvector> numbers, Mvector& sumOut, Mvector& carryOut, FHEKey& key, int index) {
    int n = numbers[0].size();
    int iter = 0;
    vector<Mvector> numbersIn = numbers, numbersOut;
    Mvector innerSumOut(n), innerCarryOut(n + 1);
    while (numbersIn.size() != 2) {
        int blocks = numbersIn.size() / 3;
        for (int i = 0; i < blocks*3; i += 3) {
            ThreeForTwoZ2(numbersIn[i], numbersIn[i + 1], numbersIn[i + 2], innerSumOut, innerCarryOut, key, index + iter);
            numbersOut.push_back(innerSumOut);
            numbersOut.push_back(innerCarryOut);
        }
        for (int i = blocks * 3; i < numbersIn.size(); i++) {
            numbersOut.push_back(Refresh(numbersIn[i], key, index + iter));
        }
        numbersIn = numbersOut;
        numbersOut.clear();
        iter++;
    }
    sumOut = numbersIn[0];
    carryOut = numbersIn[1];
    return index + iter;
}

int MultZ2(Mvector c1, Mvector c2, FHEKey& key, int index) {
    vector<Mvector> numbers;
    int length = c1.size() * c2.size();
    for (int i = 0; i < c2.size(); i++) {
        Mvector number(length);
        for (int j = 0; j < length; j++) {
            if (j < i) {
                //number[i] = 
            }
        }
    }
}

Matrix<Poly<d, Z>> Search(Mvector data, Mvector word, FHEKey& key, int& index) {
    Mvector blocksIn, blocksOut;
    int n = data.size();
    int m = word.size();
    int iter = index;
    for (int i = 0; i < n - m; i++) {
        cout << "Block " << i << endl;
        Mvector innerBlocksIn, innerBlocksOut;
        Poly<d, Z>::push_q(key.q[index]);
        for (int j = 0; j < m; j++) {
            innerBlocksIn.push_back(Refresh(data[i + j] + word[j], key, index));
        }
        iter = index + 1;
        while (innerBlocksIn.size() != 1) {
            int blocks = innerBlocksIn.size() / 2;
            Poly<d, Z>::push_q(key.q[iter]);
            for (int i = 0; i < blocks * 2; i += 2) {
                innerBlocksOut.push_back(Refresh(innerBlocksIn[i].outer(innerBlocksIn[i+1]).addToFirstRow(innerBlocksIn[i] + innerBlocksIn[i+1]), key, iter));
            }
            for (int i = blocks * 2; i < innerBlocksIn.size(); i++) {
                innerBlocksOut.push_back(Refresh(innerBlocksIn[i], key, iter));
            }
            innerBlocksIn = innerBlocksOut;
            innerBlocksOut.clear();
            iter++;
        }
        blocksIn.push_back(innerBlocksIn[0]);
    }
    while (blocksIn.size() != 1) {
        int blocks = blocksIn.size() / 2;
        Poly<d, Z>::push_q(key.q[iter]);
        for (int i = 0; i < blocks * 2; i += 2) {
            blocksOut.push_back(Refresh(blocksIn[i].outer(blocksIn[i + 1]), key, iter));
        }
        for (int i = blocks * 2; i < blocksIn.size(); i++) {
            blocksOut.push_back(Refresh(blocksIn[i], key, iter));
        }
        blocksIn = blocksOut;
        blocksOut.clear();
        iter++;
    }
    index = iter;
    return blocksIn[0];
}

void FHEKeyGen(FHEKey& key) {
    for (int i = 0; i < key.q.size(); i++) {
        Poly<d, Z>::push_q(key.q[i]);
        int N = ceil((2 * n + 1) * log2((double)key.q[i]));
        key.s.push_back(SecretKeyGen(n));
        key.A.push_back(PublicKeyGen(key.s[i], key.q[i], N));
        if (i > 0) {
            Poly<d, Z>::push_q(key.q[i - 1]);
            auto sp = key.s[i - 1].outer(key.s[i - 1]) % key.q[i - 1];
            auto spp = BitDecomp(sp, key.q[i - 1]);
            //Poly<d, Z>::push_q(qs[i]);
            key.tau.push_back(SwitchKeyGen(sp, key.s[i], key.q[i - 1]));
        }
    }
}

Matrix<Poly<d, Z>> Refresh(Matrix<Poly<d, Z>> c, FHEKey& key, int index) {
    Poly<d, Z>::push_q(key.q[index]);
    auto c2 = SwitchKey(c, key.tau[index], key.q[index]);
    auto c3 = Scale(c2, key.q[index], key.q[index + 1], 2);
    //Poly<d, Z>::push_q(key.q[index + 1]);
    Poly<d, Z>::pop_q();
    return c3;
}

Mvector Refresh(Mvector c, FHEKey& key, int index) {
    Mvector res;
    for (int i = 0; i < c.size(); i++) {
        res.push_back(Refresh(c[i], key, index));
    }
    return res;
}

Matrix<Poly<d, Z>> SecretKeyGen(int n) {
    Matrix<Poly<d, Z>> s(n + 1, 1);
    s[0] = Poly<d, Z>::One();
    for (int i = 0; i < n; i++) {
        s[i + 1] = Poly<d, Z>::Chi();
    }
    return s;
}

Matrix<Poly<d, Z>> PublicKeyGen(Matrix<Poly<d, Z>>& s, Z q, int N) {
    int n = s.size() - 1;
    Matrix<Poly<d, Z>> Ap = Matrix<Poly<d, Z>>::Uniform(q, N, n);
    Matrix<Poly<d, Z>> e = Matrix<Poly<d, Z>>::Chi(N, 1);
    Matrix<Poly<d, Z>> b = Ap * s.range(1, 0, -1, -1) + 2*e;
    Matrix<Poly<d, Z>> A = b | -Ap;
    return A;
}

Matrix<Poly<d, Z>> Enc(Matrix<Poly<d, Z>>& A, Poly<d, Z>& m) {
    int N = A.rows();
    int n = A.cols();
    Matrix<Poly<d, Z>> mm(n, 1);
    mm[0] = m;
    Matrix<Poly<d, Z>> c = (mm + A.randomRowSum().transpose());
    return c;
}

Matrix<Poly<d, Z>> Enc(Poly<d, Z>& m, FHEKey key, int index) {
    Poly<d, Z>::push_q(key.q[index]);
    auto c = Enc(key.A[index], m);
    Poly<d, Z>::pop_q();
    return c;
}

Matrix<Poly<d, Z>> EncZero(FHEKey key, int index) {
    Poly<d, Z>::push_q(key.q[index]);
    Poly<d, Z> m;
    auto c = Enc(key.A[index], m);
    Poly<d, Z>::pop_q();
    return c;
}

Matrix<Poly<d, Z>> GetZero(FHEKey& key, int index) {
    int n = key.A[index].cols();
    Matrix<Poly<d, Z>> mm(n, 1);
    return mm;
}

inline Poly<d, Z> Dec(Matrix<Poly<d, Z>> c, Matrix<Poly<d, Z>> s, Z q) {
    Poly<d, Z>::push_q(q);
    auto res = c.dot(s) % q % 2;
    Poly<d, Z>::pop_q();
    return res;
}

Poly<d, Z> Dec(Matrix<Poly<d, Z>> c, FHEKey& key, int index) {
    Poly<d, Z>::push_q(key.q[index]);
    auto res = c.dot(key.s[index]) % key.q[index] % 2;
    Poly<d, Z>::pop_q();
    return res;
}

Matrix<Poly<d, Z>> BitDecomp(Matrix<Poly<d, Z>> x, Z q) {
    int n = x.size();
    int places = ceil(log2((double)q));
    //cout << "places= " << places << endl;
    Matrix<Poly<d, Z>> res(n*places, 1);
    for (int i = 0; i < n; i++) {
        Poly<d, Z> temp = x[i].modPos(q);
        //cout << "x= " << temp << endl;
        for (int j = 0; j < places; j++) {
            res[j * n + i] = temp % 2;
            temp /= 2;
        }
    }
    return res;
}

Matrix<Poly<d, Z>> PowersOf2(Matrix<Poly<d, Z>> x, Z q) {
    int n = x.size();
    int places = ceil(log2((double)q));
    Matrix<Poly<d, Z>> res(n * places, 1);
    for (int i = 0; i < n; i++) {
        for (int j = 0; j < places; j++) {
            Z power = 1;
            power = power << j;
            res[j*n + i] = power * x[i];
        }
    }
    return res;
}

Matrix<Poly<d, Z>> SwitchKeyGen(Matrix<Poly<d, Z>> s1, Matrix<Poly<d, Z>> s2, Z q) {
    int N = s1.size() * ceil(log2((double)q));
    Matrix<Poly<d, Z>> A = PublicKeyGen(s2, q, N);
    Matrix<Poly<d, Z>> P = PowersOf2(s1, q);
    for (int i = 0; i < N; i++) {
        A(i, 0) += P[i];
    }
    return A.transpose();
}

Matrix<Poly<d, Z>> SwitchKey(Matrix<Poly<d, Z>> c, Matrix<Poly<d, Z>> B, Z q) {
    int size = (n + 1) * (n + 1);
    Matrix<Poly<d, Z>> c2(n + 1, n + 1);
    if (c.size() < (n + 1) * (n + 1)) {
        for (int i = 0; i < c.size(); i++) c2[i] = c[i];
    }
    else {
        c2 = c;
    }
    return B*BitDecomp(c2, q);
}

Matrix<Poly<d, Z>> Scale(Matrix<Poly<d, Z>> x, Z q, Z p, int r) {
    int n = x.rows();
    int m = x.cols();
    Matrix<Poly<d, Z>> res(n, m);
    Poly<d, Z>::push_q(q);
    for (int i = 0; i < n * m; i++) {
        res[i] = q - p * x[i];
    }
    Poly<d, Z>::pop_q();
    Poly<d, Z>::push_q(p);
    Z qinv = Montgomery::inv(q % p);
    for (int i = 0; i < n * m; i++) {
        res[i] = qinv * res[i];
        for (int j = 0; j < d; j++) {
            res[i][j] += x[i][j] % r - res[i][j] % r;
        }
    }
    Poly<d, Z>::pop_q();
    return res;
}

double Error(Matrix<Poly<d, Z>> c, Matrix<Poly<d, Z>> s, Z q) {
    //Poly<d, Z> m = (c.dot(s) % q) % 2;
    //cout << (c.dot(s)) % q << endl;
    return (double)((c.dot(s)) % q).maxAbsCoeff() / (double)q;
}

bool isPrimeLog(Z p) {
    if (p % 2 == 0) return false;
    for (int i = 3; i < min(sqrt((double)p), 100); i += 2) {
        if (p % i == 0) return false;
    }
    return true;
}

bool MillerRabin(Z p) {
    UniformDist dist(p);
    Montgomery::set_mod(p);
    Z d = p - 1;
    int r = 0;
    while (d % 2 == 0) {
        d >>= 1;
        r += 1;
    }
    bool cont = false;
    for (int i = 0; i < 64; i++) {
        Z a = dist.get();
        if (a == p - 1) a--;
        else if (a == 0 || a == 1) a = 2;
        Z x = Montgomery::modPow(a, d);
        if (x == 1 || x == p - 1) continue;
        cont = false;
        for (int j = 0; j < r - 1; j++) {
            x = Montgomery::modPow(x, 2);
            if (x == p - 1) {
                cont = true;
                break;
            }
        }
        if (cont) continue;
        return false;
    }
    return true;
}

bool isPrime(Z p) {
    if (!isPrimeLog(p)) return false;
    else return MillerRabin(p);
}

Z firstPrimeAbove(Z n) {
    if (n <= 1) return 2;
    if (n % 2 == 0) n++;
    while (!isPrimeLog(n) || !MillerRabin(n)) {
        n += 2;
    }
    return n;
}

//Calcula el primer primer més gran que n de la forma a*lcm(m, 2^{d+1}) + 1, amb a primer.
Z firstPrimeAboveWithConditions(Z n) {
    if (n <= 1) return 2;
    Z pow2 = 1, lcm = phi_m;
    while (pow2 < 2 * d) {
        pow2 <<= 1;
        if (lcm % 2 == 0) lcm >>= 1;
    }
    lcm = lcm * pow2;
    Z a = n / lcm + 1;
    if (a % 2 == 0) a++;
    Z p = a * lcm + 1;
    while (1) {
        p = a * lcm + 1;
        if (!isPrime(a) || !isPrime(p)) {
            a += 2;
        }
        else {
            // Trobem una arrel primitiva de la unitat mòdul p
            vector<Z> mfactors;
            for (int i = 3; i < sqrt(phi_m); i += 2) {
                if (isPrime(i) && (phi_m % i == 0)) mfactors.push_back(i);
            }
            Montgomery::set_mod(p);
            for (int i = 3; i < p; i++) {
                if (Montgomery::modPow(i, (p-1) / a) != 1 && Montgomery::modPow(i, (p-1) / 2) != 1) {
                    bool isPrimRoot = true;
                    for (int j = 0; j < mfactors.size(); j++) {
                        if (Montgomery::modPow(i, (p - 1) / mfactors[j]) == 1) {
                            isPrimRoot = false;
                            break;
                        }
                    }
                    if (isPrimRoot) {
                        primroots[p] = Montgomery::modPow(i, (p-1) / pow2);
                        break;
                    }
                }
            }
            break;
        }
    }
    return p;
}

Z gcdExtended(Z a, Z b, Z& x, Z& y){
    if (a == 0){
        x = 0;
        y = 1;
        return b;
    }

    Z x1, y1;
    Z gcd = gcdExtended(b % a, a, x1, y1);
    x = y1 - (b / a) * x1;
    y = x1;
    return gcd;
}
